/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 1998
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#import "FormFillController.h"
#import "CHBrowserView.h"
#import "KeychainService.h"

#include "nsString.h"
#include "NSString+Utils.h"
#include "nsPIDOMWindow.h"
#include "nsIDocument.h"
#include "nsIDOMWindow.h"
#include "nsIDOMDocumentEvent.h"
#include "nsIPrivateDOMEvent.h"
#include "nsIDOMEventListener.h"
#include "nsIDOMEventTarget.h"
#include "nsIDOMEvent.h"
#include "nsIDOMHTMLFormElement.h"
#include "nsIDOMNSHTMLInputElement.h"
#include "nsIDOMKeyEvent.h"
#include "nsIPresShell.h"
#include "nsIFrame.h"
#include "nsRect.h"

#include "nsIAutoCompleteResults.h"
#include "nsIAutoCompleteListener.h"

static const int kMaxRows = 6;

@interface FormFillController(Private)

// Listener and autocomplete initialization and cleanup
- (void) addWindowListeners:(nsIDOMWindow*)aWindow;
- (void) removeWindowListeners:(nsIDOMWindow*)aWindow;
- (void) startControllingInput:(nsIDOMHTMLInputElement*)aInput;
- (void) stopControllingInput;

// Popup window management
- (BOOL) isPopupOpen;
- (void) openPopup;
- (void) closePopup;
- (void) selectRowBy:(int)aRows;

// Autocomplete methods
- (void) startSearch:(nsAString&)value;
- (void) dataReady:(nsIAutoCompleteResults*)aResults status:(AutoCompleteStatus)aStatus;
- (void) doAutoComplete;
- (void) didAutoComplete;

// Event handlers
- (void) focus:(nsIDOMEvent*)aEvent;
- (void) blur:(nsIDOMEvent*)aEvent;
- (void) unload:(nsIDOMEvent*)aEvent;
- (void) submit:(nsIDOMEvent*)aEvent;
- (void) input:(nsIDOMEvent*)aEvent;
- (void) keyPress:(nsIDOMEvent*)aEvent;
- (BOOL) handleKeyNavigation:(int)aKey;

// Utility methods
- (BOOL) caretIsAtEndOfLine;
- (nsIPresShell*) getPresShellForElement:(nsIDOMElement*)aElement;
- (void) scrollElementIntoView:(nsIDOMElement*)aElement;
- (nsRect) getScreenOrigin:(nsIDOMElement*)aElement;

@end

class FormFillListener : public nsIDOMEventListener, 
                         public nsIAutoCompleteListener
{ 
public:
  NS_DECL_ISUPPORTS

  // nsIAutoCompleteListener functions

  NS_IMETHODIMP OnStatus(const PRUnichar* aText) { return NS_OK; }
  NS_IMETHODIMP SetParam(nsISupports *aParam) { return NS_OK; }
  NS_IMETHODIMP GetParam(nsISupports **aParam) { return NS_OK; }

  NS_IMETHODIMP OnAutoComplete(nsIAutoCompleteResults *aResults, 
                               AutoCompleteStatus aStatus)
  {
    [[FormFillController instance] dataReady:aResults status:aStatus];
    return NS_OK;
  }

  // nsIDOMEventListener functions

  // HandleEvent
  // Calls the methods in FormFillController to handle DOM events
  NS_IMETHODIMP FormFillListener::HandleEvent(nsIDOMEvent* aEvent)
  {
    nsAutoString type;
    aEvent->GetType(type);

    if (type.Equals(NS_LITERAL_STRING("focus")))
      [[FormFillController instance] focus:aEvent];
    else if (type.Equals(NS_LITERAL_STRING("blur")))
      [[FormFillController instance] blur:aEvent];
    else if (type.Equals(NS_LITERAL_STRING("unload")))
      [[FormFillController instance] unload:aEvent];
    else if (type.Equals(NS_LITERAL_STRING("submit")))
      [[FormFillController instance] submit:aEvent];
    else if (type.Equals(NS_LITERAL_STRING("input")))
      [[FormFillController instance] input:aEvent];
    else if (type.Equals(NS_LITERAL_STRING("keypress")))
      [[FormFillController instance] keyPress:aEvent];
    return NS_OK; 
  }
};

NS_IMPL_ISUPPORTS2(FormFillListener, nsIDOMEventListener, nsIAutoCompleteListener)

@implementation FormFillController

static FormFillController *sInstance = nil;

+ (FormFillController*)instance
{
  return sInstance ? sInstance : sInstance = [[self alloc] init];
}

- (id)init
{
  if ((self = [super init])) {
    // mListener captures DOM and auto complete events
    mListener = new FormFillListener;
    NS_ADDREF(mListener);

    // Initialize the password fill session.
    // History form fill session can also be added when someone writes it
    mKeychainSession = new KeychainAutoCompleteSession;
    NS_ADDREF(mKeychainSession);

    mBrowserViews = [[NSMutableArray alloc] init];
    mPopups = [[NSMutableArray alloc] init];

    mSearchString = [[NSMutableString alloc] init];
  }

  return self;
}

- (void)dealloc
{
  [mSearchString release];

  // Remove ourselves as a focus listener from all cached views
  int count = [mBrowserViews count];
  for (int i = 0; i < count; ++i) {
    CHBrowserView* browser = [mBrowserViews objectAtIndex:i];
    nsCOMPtr<nsIDOMWindow> domWindow = [browser getContentWindow];
    if (domWindow)
      [self removeWindowListeners:domWindow];
  }

  NS_IF_RELEASE(mResults);
  NS_RELEASE(mKeychainSession);
  NS_RELEASE(mListener);

  [mBrowserViews release];
  [mPopups release];

  [super dealloc];
}

// listeners ////////////////////////////

- (void)attachToBrowser:(CHBrowserView*)aBrowser withPopup:(FormFillPopup*)aPopup
{
  if (!aBrowser || !aPopup)
    return;
  
  [mBrowserViews addObject:aBrowser];
  [mPopups addObject:aPopup];

  // Listen for focus events on the domWindow of the browser view
  nsCOMPtr<nsIDOMWindow> domWindow = [aBrowser getContentWindow];
  if (domWindow)
    [self addWindowListeners:domWindow];
}

- (void)detachFromBrowser:(CHBrowserView*)aBrowser
{
  PRInt32 index = [mBrowserViews indexOfObject:aBrowser];
  if (index < 0)
    return;

  // Stop listening for focus events on the domWindow of the docShell
  nsCOMPtr<nsIDOMWindow> domWindow = [aBrowser getContentWindow];
  if (domWindow)
    [self removeWindowListeners:domWindow];
    
  [mBrowserViews removeObjectAtIndex:index];
  [mPopups removeObjectAtIndex:index];
}

- (void)addWindowListeners:(nsIDOMWindow*)aWindow
{
  if (!aWindow)            
    return;
  
  nsCOMPtr<nsPIDOMWindow> privateDOMWindow = do_QueryInterface(aWindow);
  if (!privateDOMWindow)
    return;

  nsPIDOMEventTarget* chromeEventHandler = nsnull;
  chromeEventHandler = privateDOMWindow->GetChromeEventHandler();

  nsCOMPtr<nsIDOMEventTarget> target = do_QueryInterface(chromeEventHandler);

  if (!target)
    return;

  target->AddEventListener(NS_LITERAL_STRING("focus"),
                           static_cast<nsIDOMEventListener*>(mListener),
                           PR_TRUE);

  target->AddEventListener(NS_LITERAL_STRING("blur"),
                           static_cast<nsIDOMEventListener*>(mListener),
                           PR_TRUE);

  target->AddEventListener(NS_LITERAL_STRING("unload"),
                           static_cast<nsIDOMEventListener*>(mListener),
                           PR_TRUE);

  target->AddEventListener(NS_LITERAL_STRING("submit"),
                           static_cast<nsIDOMEventListener*>(mListener),
                           PR_TRUE);
                                                  
  target->AddEventListener(NS_LITERAL_STRING("input"),
                           static_cast<nsIDOMEventListener*>(mListener),
                           PR_TRUE);

  target->AddEventListener(NS_LITERAL_STRING("keypress"),
                           static_cast<nsIDOMEventListener*>(mListener),
                           PR_TRUE);
}

- (void)removeWindowListeners:(nsIDOMWindow*)aWindow
{
  [self stopControllingInput];
  
  nsCOMPtr<nsPIDOMWindow> privateDOMWindow = do_QueryInterface(aWindow);
  if (!privateDOMWindow)
    return;

  nsPIDOMEventTarget* chromeEventHandler = nsnull;
  chromeEventHandler = privateDOMWindow->GetChromeEventHandler();

  nsCOMPtr<nsIDOMEventTarget> target = do_QueryInterface(chromeEventHandler);

  if (!target)
    return;

  target->RemoveEventListener(NS_LITERAL_STRING("focus"),
                              static_cast<nsIDOMEventListener*>(mListener),
                              PR_TRUE);

  target->RemoveEventListener(NS_LITERAL_STRING("blur"),
                              static_cast<nsIDOMEventListener*>(mListener),
                              PR_TRUE);

  target->RemoveEventListener(NS_LITERAL_STRING("unload"),
                              static_cast<nsIDOMEventListener*>(mListener),
                              PR_TRUE);

  target->RemoveEventListener(NS_LITERAL_STRING("submit"),
                              static_cast<nsIDOMEventListener*>(mListener),
                              PR_TRUE);

  target->RemoveEventListener(NS_LITERAL_STRING("input"),
                              static_cast<nsIDOMEventListener*>(mListener),
                              PR_TRUE);

  target->RemoveEventListener(NS_LITERAL_STRING("keypress"),
                              static_cast<nsIDOMEventListener*>(mListener),
                              PR_TRUE);
}

- (void)startControllingInput:(nsIDOMHTMLInputElement*)aInput
{
  // Make sure we're not still attached to an input
  [self stopControllingInput];

  // Set the autocomplete session for this input.  Currently, only password autocomplete
  // but other sessions like form fill history can be added here.
  mUsernameFillEnabled = mKeychainSession->AttachToInput(aInput);

  // If this is not a valid auto-complete input, we can return now
  if (!mUsernameFillEnabled)
    return;

  // Find the currently focused CHBrowserView to get the correct popup
  nsCOMPtr<nsIDOMDocument> domDoc;
  aInput->GetOwnerDocument(getter_AddRefs(domDoc));
  nsCOMPtr<nsIDocument> doc = do_QueryInterface(domDoc);
  if (!doc)
    return;

  CHBrowserView* browser = [CHBrowserView browserViewFromDOMWindow:doc->GetWindow()];

  int index = [mBrowserViews indexOfObject:browser];
  if (index < 0)
    return;
  
  // Cache the popup and view for the focused docShell
  mFocusedPopup = [mPopups objectAtIndex:index];
  mFocusedView = browser;
  mFocusedInput = aInput;  

  // Set the search string to the value of the input
  nsAutoString value;
  mFocusedInput->GetValue(value);
  [mSearchString setString:[NSString stringWith_nsAString:value]];
}

- (void)stopControllingInput
{
  if ([self isPopupOpen])
    [self closePopup];

  [mSearchString setString:@""];

  mFocusedInput = nsnull;
  mFocusedPopup = nil;
  mFocusedView = nil;

  // Stop sending search requests to auto-form fill.  
  mUsernameFillEnabled = FALSE;

  NS_IF_RELEASE(mResults);
  mResults = nsnull;
  [mFocusedPopup setResults:nsnull];
}

// popup management ////////////////////////////

- (BOOL)isPopupOpen
{
  return (mFocusedPopup) ? [mFocusedPopup isPopupOpen] : FALSE;
}

- (void)openPopup
{
  // Only open popup if it exists and is not already open
  if (!mFocusedPopup || [self isPopupOpen])
    return;

  // Make sure input field is visible before showing popup
  [self scrollElementIntoView:mFocusedInput];

  nsRect inputRect = [self getScreenOrigin:mFocusedInput];

  // Set the lower left corner of the input element as the upper left corner
  // of the popup and the width of the element as the width of the popup
  NSPoint point = NSMakePoint(inputRect.x, inputRect.y + inputRect.height);
  [mFocusedPopup openPopup:mFocusedView orig:point width:inputRect.width];
}

- (void)closePopup
{
  if (mFocusedPopup) {
    // Deselecting the row prevents a flash when popup is opened and default is selected
    [mFocusedPopup selectRow:-1];
    [mFocusedPopup closePopup];
  }
}

- (void)selectRowBy:(int)aRows
{
  int row = [mFocusedPopup selectedRow] + aRows;

  // pin result at top row
  if (row < 0)
    row = 0;
    
  // pin result at bottom row
  int numrows = [mFocusedPopup rowCount];
  if (row >= numrows)
    row = numrows-1;

  [mFocusedPopup selectRow:row];
}

//
// popupSelected
//
// Called when a new item in the popup window is selected, either by mouse click
// or keyboard movement.  Like Safari, we want to set the form field to the 
// value of the selected item and send an autocomplete event for any listeners
- (void)popupSelected
{
  if (!mFocusedPopup || ![self isPopupOpen])
    return;

  int row = [mFocusedPopup selectedRow];
  if (row < 0)
    return;

  nsAutoString value;
  [[mFocusedPopup resultForRow:row] assignTo_nsAString:value];
  mFocusedInput->SetValue(value);

  // Auto-fill the input like Safari where untyped letters are selected
  nsCOMPtr<nsIDOMNSHTMLInputElement> nsInput = do_QueryInterface(mFocusedInput);
  if (nsInput) {
    int searchLength = [mSearchString length];
    nsInput->SetSelectionStart(searchLength);
  }

  // Send an autocomplete DOM event any time auto fill is done
  [self didAutoComplete];
}

// auto complete ////////////////////////////
 
- (void) startSearch:(nsAString&)aString;
{
  [mSearchString setString:[NSString stringWith_nsAString:aString]];
 
  // Check if password autocomplete is enabled.  
  // Form history autocomplete can be added here.
  if (mUsernameFillEnabled)
    mKeychainSession->OnStartLookup([mSearchString createNewUnicodeBuffer], mResults, mListener);
} 

- (void) dataReady:(nsIAutoCompleteResults*)aResults status:(AutoCompleteStatus)aStatus
{
  NS_IF_RELEASE(mResults);
  mResults = nsnull;

  if (aStatus == nsIAutoCompleteStatus::matchFound) {
    mResults = aResults;
    NS_IF_ADDREF(mResults);

    [mFocusedPopup setResults:mResults];

    // Open the popup if more than one result is returned. 
    // Also require mCompleteResult since it prevents a backspace from opening popup
    // even though we want to search on a backspace to get the larger result set
    if ([mFocusedPopup rowCount] > 1 && mCompleteResult) {
      [self openPopup];
    } 
    else 
      [self closePopup];

    // Prevents backspace from auto-completing the result we just erased
    if (mCompleteResult)
      [self doAutoComplete];
  }
  else {
    [mFocusedPopup setResults:nil];
    [self closePopup];
  }
}

// XXX Put comment here
- (void) doAutoComplete
{
  if (!mResults)
    return;

  nsCOMPtr<nsISupportsArray> items;
  mResults->GetItems(getter_AddRefs(items));

  // Select the default if it's available
  // Otherwise, select the first username in the list
  PRInt32 defaultIndex;
  mResults->GetDefaultItemIndex(&defaultIndex);
  if (defaultIndex < 0)
    defaultIndex = 0;

  nsCOMPtr<nsISupports> itemSupports;
  itemSupports = dont_AddRef(items->ElementAt(defaultIndex));

  nsCOMPtr<nsIAutoCompleteItem> item = do_QueryInterface(itemSupports);
  if (!item)
    return;

  nsAutoString value;
  item->GetValue(value);
  mFocusedInput->SetValue(value);

  // Auto-fill the input like Safari where untyped letters are selected
  nsCOMPtr<nsIDOMNSHTMLInputElement> nsInput = do_QueryInterface(mFocusedInput);
  if (nsInput) {
    int searchLength = [mSearchString length];
    nsInput->SetSelectionStart(searchLength);
  }

  // Select the default entry in the popup
  if ([self isPopupOpen])
    [mFocusedPopup selectRow:defaultIndex];

  // Send a DOM event any time auto fill is done
  [self didAutoComplete];
}

// XXX Add comment here
- (void) didAutoComplete
{
  if (!mFocusedInput)
    return;

  nsCOMPtr<nsIDOMDocument> domDoc;
  mFocusedInput->GetOwnerDocument(getter_AddRefs(domDoc));

  nsCOMPtr<nsIDOMDocumentEvent> doc = do_QueryInterface(domDoc);
  if (!doc)
    return;
  
  nsCOMPtr<nsIDOMEvent> event;
  doc->CreateEvent(NS_LITERAL_STRING("Events"), getter_AddRefs(event));
  nsCOMPtr<nsIPrivateDOMEvent> privateEvent = do_QueryInterface(event);
  if (!privateEvent)
    return;
    
  event->InitEvent(NS_LITERAL_STRING("DOMAutoComplete"), PR_TRUE, PR_TRUE);

  // XXXjst: We mark this event as a trusted event, it's up to the
  // callers of this to ensure that it's only called from trusted
  // code.
  privateEvent->SetTrusted(PR_TRUE);

  nsCOMPtr<nsIDOMEventTarget> targ = do_QueryInterface(mFocusedInput);

  PRBool defaultActionEnabled;
  targ->DispatchEvent(event, &defaultActionEnabled);
}

// event handlers ////////////////////////////

//
// focus
//
// When a new element is selected, check whether we allow autocomplete
// If autocomplete is allowed, start controlling the input to that element
//
- (void)focus:(nsIDOMEvent*)aEvent
{
  nsCOMPtr<nsIDOMEventTarget> target;
  aEvent->GetTarget(getter_AddRefs(target));

  nsCOMPtr<nsIDOMHTMLInputElement> input = do_QueryInterface(target);
  if (!input)
    return;

  nsAutoString type;
  input->GetType(type);

  PRBool isReadOnly = PR_FALSE;
  input->GetReadOnly(&isReadOnly);

  nsAutoString autocomplete;
  input->GetAttribute(NS_LITERAL_STRING("autocomplete"), autocomplete);
  if (type.LowerCaseEqualsLiteral("text") && !isReadOnly &&
      !autocomplete.LowerCaseEqualsLiteral("off")) {

    nsCOMPtr<nsIDOMHTMLFormElement> form;
    input->GetForm(getter_AddRefs(form));
    if (form)
      form->GetAttribute(NS_LITERAL_STRING("autocomplete"), autocomplete);

    if (!form || !autocomplete.LowerCaseEqualsLiteral("off"))
      [self startControllingInput:input];
  }
}

//
// blur
//
// Anytime focus moves from an element, stop autocompleting it until next
// input element is focused
//
- (void)blur:(nsIDOMEvent*)aEvent
{
  if (mFocusedInput)
    [self stopControllingInput];
}

//
// unload
//
// Stop autocompleting on a page if it is unloaded
//
- (void)unload:(nsIDOMEvent*)aEvent
{
  if (mFocusedInput) {
    nsCOMPtr<nsIDOMEventTarget> target;
    aEvent->GetTarget(getter_AddRefs(target));

    nsCOMPtr<nsIDOMDocument> eventDoc = do_QueryInterface(target);
    nsCOMPtr<nsIDOMDocument> inputDoc;
    mFocusedInput->GetOwnerDocument(getter_AddRefs(inputDoc));

    if (eventDoc == inputDoc)
      [self stopControllingInput];
  }
}

//
// submit
//
// If a form is submitted, stop autocompleting
//
- (void)submit:(nsIDOMEvent*)aEvent
{
  if (mFocusedInput)
    [self stopControllingInput];
}

//
// input
//
// If a HTML input box is being controlled, do a search when input occurs
//
- (void)input:(nsIDOMEvent*)aEvent
{
  if (mFocusedInput) {
    nsAutoString value;
    mFocusedInput->GetValue(value);
    [self startSearch:value];    
  }
}

//
// keyPress
//
// This is triggered when a key is pressed but before an 'input' is triggered.
// It also handles non-input keys like arrow keys
//
- (void)keyPress:(nsIDOMEvent*)aEvent
{
  if (!mFocusedInput)
    return;

  nsCOMPtr<nsIDOMKeyEvent> keyEvent = do_QueryInterface(aEvent);
  if (!keyEvent)
    return;

  // By default, allow keystroke to continue to the next listener
  BOOL cancel = FALSE;

  // By default, autocomplete on keystrokes that later trigger 'input' events
  mCompleteResult = TRUE;

  PRUint32 k;
  keyEvent->GetKeyCode(&k);
  switch (k) {
  case nsIDOMKeyEvent::DOM_VK_UP:
  case nsIDOMKeyEvent::DOM_VK_DOWN:
  case nsIDOMKeyEvent::DOM_VK_PAGE_UP:
  case nsIDOMKeyEvent::DOM_VK_PAGE_DOWN:
    cancel = [self handleKeyNavigation:k];
    break;
  case nsIDOMKeyEvent::DOM_VK_ESCAPE:
  case nsIDOMKeyEvent::DOM_VK_RETURN:
    cancel = [self isPopupOpen];
    [self closePopup];
    break;
  case nsIDOMKeyEvent::DOM_VK_BACK_SPACE:
  case nsIDOMKeyEvent::DOM_VK_DELETE:
    // Don't allow autocomplete of 'input' event if it's due to a back space or delete
    mCompleteResult = FALSE;
    break;
  }

  if (cancel) {
    aEvent->StopPropagation();
    aEvent->PreventDefault();
  }
}

// XXX Comment return value
- (BOOL)handleKeyNavigation:(int)aKey
{  
  switch (aKey) {
  case nsIDOMKeyEvent::DOM_VK_UP:
    if ([self isPopupOpen]) {
      [self selectRowBy:-1];
      [self popupSelected];
      return TRUE;
    }
    break;
  case nsIDOMKeyEvent::DOM_VK_DOWN:
    if ([self isPopupOpen]) {
      [self selectRowBy:1];
      [self popupSelected];
      return TRUE;
    }
    else if ([self caretIsAtEndOfLine]) {
      nsAutoString value;
      mFocusedInput->GetValue(value);
      [self startSearch:value];
      return TRUE;
    }
    break;
  case nsIDOMKeyEvent::DOM_VK_PAGE_UP:
    if ([self isPopupOpen]) {
      [self selectRowBy:-kMaxRows];
      [self popupSelected];
      return TRUE;
    }
    break;
  case nsIDOMKeyEvent::DOM_VK_PAGE_DOWN:
    if ([self isPopupOpen]) {
      [self selectRowBy:kMaxRows];
      [self popupSelected];
      return TRUE;
    }
    break;
  }

  return FALSE;
}

// utility functions ////////////////////////////

- (BOOL) caretIsAtEndOfLine
{
  if (!mFocusedInput)
    return FALSE;
  
  nsCOMPtr<nsIDOMNSHTMLInputElement> nsInput = do_QueryInterface(mFocusedInput);
  if (!nsInput)
    return FALSE;
 
  PRInt32 selectStart;
  nsInput->GetSelectionStart(&selectStart);

  PRInt32 textLength;
  nsInput->GetTextLength(&textLength);

  return (selectStart == textLength);
}

- (nsIPresShell*)getPresShellForElement:(nsIDOMElement*)aElement
{
  nsCOMPtr<nsIContent> content = do_QueryInterface(aElement);
  NS_ENSURE_TRUE(content, nsnull);
    
  nsCOMPtr<nsIDocument> doc = content->GetDocument();
  NS_ENSURE_TRUE(doc, nsnull);

  return doc->GetPrimaryShell();
}

- (void)scrollElementIntoView:(nsIDOMElement*)aElement
{
  nsIPresShell* presShell = [self getPresShellForElement:aElement];
  if (!presShell)
    return;

  nsCOMPtr<nsIContent> content = do_QueryInterface(aElement);
  if (!content)
    return;
    
  presShell->ScrollContentIntoView(content, NS_PRESSHELL_SCROLL_IF_NOT_VISIBLE,
                                            NS_PRESSHELL_SCROLL_IF_NOT_VISIBLE);
}

- (nsRect)getScreenOrigin:(nsIDOMElement*)aElement
{
  nsRect rect(0,0,0,0);

  nsIPresShell* presShell = [self getPresShellForElement:aElement];

  if (presShell) {
    nsCOMPtr<nsIContent> content = do_QueryInterface(aElement);
    nsIFrame* frame = presShell->GetPrimaryFrameFor(content);
    if (frame)
      rect = frame->GetScreenRectExternal();
  }

  return rect;
}

@end
