/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 2002
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#import "NSString+Utils.h"

#import "FormFillPopup.h"
#import "FormFillController.h"
#import "NSScreen+Utils.h"
#import "CHBrowserView.h"

#include "nsIAutoCompleteResults.h"
#include "nsString.h"

@interface FormFillPopup(Private)

- (void) onRowClicked:(NSNotification *)aNote;
- (void) onResize:(NSNotification *)aNote;

@end

@implementation FormFillDataSource

- (id) resultForRow:(int)aRow columnIdentifier:(NSString *)aColumnIdentifier
{
  NSString* result = @"";

  if (!mResults)
    return result;

  nsCOMPtr<nsISupportsArray> items;
  mResults->GetItems(getter_AddRefs(items));

  nsCOMPtr<nsISupports> itemSupports = dont_AddRef(items->ElementAt(aRow));
  nsCOMPtr<nsIAutoCompleteItem> item = do_QueryInterface(itemSupports);
  if (!item)
    return result;

  nsAutoString value;
  item->GetValue(value);
  return [NSString stringWith_nsAString:value];
}

@end

@implementation FormFillPopup

- (id)init
{
  if ((self = [super init]))
  {
    // construct and configure the popup window
    mPopupWin = [[NSWindow alloc] initWithContentRect:NSMakeRect(0,0,0,0)
                styleMask:NSBorderlessWindowMask backing:NSBackingStoreBuffered defer: NO];

    [mPopupWin setReleasedWhenClosed:NO];
    [mPopupWin setHasShadow:YES];
    [mPopupWin setAlphaValue:0.9];

    // construct and configure the view
    mTableView = [[[NSTableView alloc] initWithFrame:NSMakeRect(0,0,0,0)] autorelease];
    [mTableView setIntercellSpacing:NSMakeSize(1, 2)];
    [mTableView setTarget:self];
    [mTableView setAction:@selector(onRowClicked:)];

    // create the text column (only one)
    NSTableColumn* column = [[[NSTableColumn alloc] initWithIdentifier:@"col1"] autorelease];
    [mTableView addTableColumn:column];

    mDataSource = [[FormFillDataSource alloc] init];
    [mTableView setDataSource:mDataSource];

    [mPopupWin setContentView:mTableView];
  }
  return self;
}

- (void)dealloc
{
  [mPopupWin release];
  [mDataSource release];

  [super dealloc];
}

- (BOOL)isPopupOpen
{
  return [mPopupWin isVisible];
}

- (void)openPopup:(CHBrowserView*)aBrowser orig:(NSPoint)point width:(float)width
{
  NSScreen* screen = [NSScreen screenForPoint:point];
  if (!screen)
    screen = [NSScreen mainScreen];

  if (!screen)
    return;

  // Need to convert top-left origin to bottom-left origin
  point.y = NSHeight([screen frame]) - point.y;
  // XXX check shift here, still not aligned sometimes
  // shift popup to the right to account for text box border
  point.x += 2.0;

  // set the new size
  NSRect tableViewFrame = [mTableView frame];
  // XXX check width change here, still not aligned sometimes
  // adjust width to account for text box border
  tableViewFrame.size.width = width - 3.0;
  [mTableView setFrame:tableViewFrame];
  
  // size the panel correctly
  NSSize tableSize = tableViewFrame.size;
  [mPopupWin setContentSize:tableSize];
  [mPopupWin setFrameTopLeftPoint:point];

  // show the popup
  if ([mPopupWin isVisible] == NO)
  {
    NSWindow* window = [aBrowser window];
    [window addChildWindow:mPopupWin ordered:NSWindowAbove];
    [mPopupWin orderFront:nil];
    
    // XXX Figure out to close popup when window moves or is scrolled by mouse
    //     This should ideally be done elsewhere with a Rollup listener
    [[NSNotificationCenter defaultCenter] addObserver:self
                                             selector:@selector(onResize:)
                                                 name:NSWindowDidResizeNotification
                                               object:nil];
  }
}

- (void)resizePopup
{
  // Don't resize if popup isn't visible
  if (![mPopupWin isVisible])
    return;
 
  if ([mTableView numberOfRows] == 0) {
    [self closePopup];
    return;
  }

  // The table view is resized automatically.  Set the window frame 
  // to account for any table view height change
  NSRect popupWinFrame = [mPopupWin frame];
  NSRect tableViewFrame = [mTableView frame];

  popupWinFrame.origin.y += NSHeight(popupWinFrame) - NSHeight(tableViewFrame);
  popupWinFrame.size.height = NSHeight(tableViewFrame);
  
  [mPopupWin setFrame:popupWinFrame display:YES];
}

- (void)closePopup
{
  // we can get -closePopup even if we didn't show it
  if ([mPopupWin isVisible])
  {
    [[mPopupWin parentWindow] removeChildWindow:mPopupWin];
    [mPopupWin orderOut:nil];
    [[NSNotificationCenter defaultCenter] removeObserver:self];
  }
}

- (void) onRowClicked:(NSNotification *)aNote
{
  [[FormFillController instance] popupSelected];
  [self closePopup];
}

// XXX - see comment above
- (void) onResize:(NSNotification *)aNote
{
  [self closePopup];
}

- (int)rowCount
{
  return [mDataSource rowCount];
}

- (void)selectRow:(int)index
{
  if (index == -1)
    [mTableView deselectAll:self];
  else
    [mTableView selectRowIndexes:[NSIndexSet indexSetWithIndex:index] byExtendingSelection:FALSE];
}

- (int)selectedRow
{
  return [mTableView selectedRow];
}

- (NSString*)resultForRow:(int)index
{
  return [mDataSource resultForRow:index columnIdentifier:@"col1"];
}

- (void)setResults:(nsIAutoCompleteResults*)aResults
{
  [mDataSource setResults:aResults];

  // Update the view any time we get new data
  [mTableView noteNumberOfRowsChanged]; 
  [self resizePopup];
}

// methods for table view interaction
-(int) numberOfRowsInTableView:(NSTableView*)aTableView
{
  return [mDataSource rowCount];
}

-(id)tableView:(NSTableView*)aTableView objectValueForTableColumn:(NSTableColumn*)aTableColumn row:(int)aRowIndex
{
  return [mDataSource resultForRow:aRowIndex columnIdentifier:[aTableColumn identifier]];
}

@end
